"""
@brief   Video frame clustering based on Selective Convolutional Descriptor 
         Aggregation.

@details They key frames are selected by unsupervised clustering of 
         latent feature vectors corresponding to the video frames.
         The latent feature vector is obtained as explained in 
         Wei et al. 2017:

         "Selective Convolutional Descriptor Aggregation for
         Fine-Grained Image Retrieval"
                 
         The only difference with the paper is we use InceptionV3
         to retrieve the latent tensor (as opposed to VGG-16).

@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    1 Dec 2022.
"""

import numpy as np
import tqdm
import sklearn_extra.cluster
import imageio_ffmpeg
import skimage.measure
import scipy.ndimage
import faiss

# My imports
import videosum

def get_key_frames_scda(self, input_path):
        """
        @brief Get a list of key video frames. 
        @param[in]  input_path  Path to the video file.
        @returns a list of Numpy/BGR images, range [0.0, 1.0], dtype = np.uint8. 
    
        """
        latent_vectors = []

        # Initialise video reader
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps,
                                      pix_fmt='rgb24')
        w, h = reader.size

        # Initialise Inception network model
        model = videosum.FrechetInceptionDistance('tensor')

        # Collect feature vectors for all the frames
        print('[INFO] Collecting feature vectors for all the frames ...')
        for raw_frame in tqdm.tqdm(reader):
            # Convert video frame into a BGR OpenCV/Numpy image
            im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

            # Compute latent feature vector for this video frame
            tensor = model.get_latent_feature_tensor(im)

            # Sum tensor over channels
            aggregation_map = tensor.sum(axis=0)

            # Compute mask map
            mean = aggregation_map.mean()
            mask_map = np.zeros_like(aggregation_map)
            mask_map[aggregation_map > mean] = 1

            # Connected component analysis
            cc = skimage.measure.label(mask_map) 

            # Largest connected component
            largest = np.argmax(np.bincount(cc.flat)[1:]) + 1
            largest_cc = scipy.ndimage.binary_fill_holes(cc == largest).astype(cc.dtype)

            # Selected descriptor set
            selected_descriptor_set = []
            for i in range(8):
                for j in range(8):
                    if largest_cc[i, j]:
                        selected_descriptor_set.append(tensor[:, i, j])
            selected_descriptor_set = np.array(selected_descriptor_set)

            # Compute avg&maxPool SCDA feature vector
            avgpool = np.mean(selected_descriptor_set, axis=0)
            maxpool = np.max(selected_descriptor_set, axis=0)
            scda = np.hstack([avgpool, maxpool])

            # Add feature vector to our list
            latent_vectors.append(scda)
        print('[INFO] Done. Feature vectors computed.')

        # Compute L2 distances
        print('[INFO] Computing distance matrix ...')
        X = np.array(latent_vectors, dtype=np.float32)
        mt = getattr(faiss, 'METRIC_L2')
        l2norm = np.clip(faiss.pairwise_distances(X, X, mt), 0, None)
        print('[INFO] Done, distance matrix computed.')

        # Cluster the feature vectors
        print('[INFO] k-medoids clustering ...')
        kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=self.number_of_frames, 
            metric='precomputed',
            init='k-medoids++',
            random_state=0).fit(l2norm)
        print('[INFO] k-medoids clustering finished.')
        indices = sorted(kmedoids.medoid_indices_.tolist(), reverse=True)
        self.indices_ = [x for x in indices]
        self.labels_ = kmedoids.labels_

        # Retrieve the video frames corresponding to the cluster means
        print('[INFO] Retrieving key frames ...')
        key_frames = []
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                      pix_fmt='rgb24')
        counter = 0
        for raw_frame in tqdm.tqdm(reader):
            if counter in self.indices_:
                # Convert video frame into a BGR OpenCV/Numpy image
                im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

                # Add key frame to list
                key_frames.append(im)
                
            counter += 1
        print('[INFO] Key frames obtained.')

        return key_frames



"""
@brief  Method get_key_frames_inception() of the class VideoSummariser.
@author Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date   29 Nov 2022
"""
import numpy as np
import tqdm
import faiss
import sklearn_extra.cluster
import imageio_ffmpeg 

# My imports 
import videosum

def get_key_frames_inception(self, input_path, eps=1e-6):
        """
        @brief Get a list of key video frames. 
        @details They key frames are selected by unsupervised clustering of 
                 latent feature vectors corresponding to the video frames.
                 The latent feature vector is obtained from Inception V3
                 trained on ImageNet. The clustering method used is kmedoids.  

        @param[in]  input_path  Path to the video file.
        
        @returns a list of Numpy/BGR images, range [0.0, 1.0], dtype = np.uint8. 
        """
        latent_vectors = []

        # Initialise video reader
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                      pix_fmt='rgb24')
        w, h = reader.size

        # Initialise Inception network model
        model = videosum.FrechetInceptionDistance('vector')

        # Collect feature vectors for all the frames
        print('[INFO] Collecting feature vectors for all the frames ...')
        for raw_frame in tqdm.tqdm(reader):
            # Convert video frame into a BGR OpenCV/Numpy image
            im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

            # Compute latent feature vector for this video frame
            vec = model.get_latent_feature_vector(im)

            # Add feature vector to our list
            latent_vectors.append(vec)
        print('[INFO] Done. Feature vectors computed.')

        # Compute L2 distances
        X = np.array(latent_vectors, dtype=np.float32)
        mt = getattr(faiss, 'METRIC_L2')
        l2norm = np.clip(faiss.pairwise_distances(X, X, mt), 0, None)

        # Cluster the feature vectors
        print('[INFO] k-medoids clustering ...')
        kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=self.number_of_frames,
            metric='precomputed',
            method='pam',
            init='k-medoids++',
            random_state=0).fit(l2norm)
        self.indices_ = kmedoids.medoid_indices_
        self.labels_ = kmedoids.labels_
        print('[INFO] k-medoids clustering finished.')

        # Retrieve the video frames corresponding to the cluster means
        print('[INFO] Retrieving key frames ...')
        key_frames = []
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                      pix_fmt='rgb24')
        counter = 0
        for raw_frame in tqdm.tqdm(reader):
            if counter in self.indices_:
                # Convert video frame into a BGR OpenCV/Numpy image
                im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

                # Add key frame to list
                key_frames.append(im)
                
            counter += 1
        print('[INFO] Key frames obtained.')

        return key_frames


if __name__ == '__main__':
    raise RuntimeError('[ERROR] inception.py cannot be run as a script.')



"""
@brief Method of the class VideoSUmmariser to summarise a video as a storyboard
       of frames evenly spaced.
"""
import math
import tqdm
import numpy as np

# My imports
import videosum


def get_key_frames_time(self, input_path):
        """
        @brief Get a list of key frames from the video. The key frames are
               simply evenly spaced along the video.
        
        @param[in]  input_path  Path to the video file.

        @returns a list of Numpy/OpenCV BGR images. 
                 After this method is executed the list self.indices_ will 
                 hold the list of frame indices that represent the key frames.
        """
        key_frames = []

        # Initialise video reader
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps)

        # Get the properties of the video
        w, h = reader.size
        nframes = videosum.VideoReader.num_frames(input_path, self.fps)
        
        # We know the key frames straight away 
        self.indices_ = np.round(np.linspace(0, nframes - 1, num=self.number_of_frames)).astype(int)

        frame_count = 0
        for raw_frame in tqdm.tqdm(reader):
            if frame_count in self.indices_: 
                # Convert video frame into a BGR OpenCV/Numpy image
                im = np.frombuffer(raw_frame, 
                        dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

                # Insert video frame in our list of key frames
                key_frames.append(im)

            frame_count += 1

        # Build self.labels_ based on self.indices_
        self.labels_ = []
        for i in range(nframes):
            min_dist = np.inf
            min_idx = -1
            for j in range(len(self.indices_)):
                if np.abs(i - self.indices_[j]) < min_dist:
                    min_dist = np.abs(i - self.indices_[j])
                    min_idx = j
            self.labels_.append(min_idx)

        return key_frames


if __name__ == '__main__':
    raise RuntimeError('[ERROR] time.py cannot be run as a script.')



"""
@brief   Package to summarise a video into N frames. There are a lot of methods
         for video summarisation, and a lot of repositories in GitHub, but
         none of them seems to work out of the box. This package contains a 
         simple way of doing it.

@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    18 May 2022.
"""

import argparse
import imageio_ffmpeg 
import numpy as np
import os
import cv2
import tqdm
import time
import multiprocessing

# My imports
import videosum


def help(short_option):
    """
    @returns The string with the help information for each command line option.
    """
    help_msg = {
        '-i': 'Path to the video (required: True)',
        '-o': 'Output directory path (required: True)',
        '-n': 'Number of frames in the collage (required: True)',
        '-x': 'Width of the collage (required: True)',
        '-y': 'Height of the collage (required: True)',
        '-a': 'Algorithm for frame selection, available options: time, frechet (required: True)',
        '-t': 'Add time segmentation based on the key frame selection (required: False)',
        '-f': 'Sampling frequency in fps (required: False)', 
    }
    return help_msg[short_option]


def parse_cmdline_params():
    """@returns The argparse args object."""
    # Create command line parser
    parser = argparse.ArgumentParser(description='Easy-to-use video summarisation.')
    parser.add_argument('-i', '--input', required=True, type=str, 
                        help=help('-i'))
    parser.add_argument('-o', '--output', required=True, type=str, 
                        help=help('-o'))
    parser.add_argument('-n', '--nframes', required=True, type=int, 
                        help=help('-n'))
    parser.add_argument('-x', '--width', required=True, type=int, 
                        help=help('-x'))
    parser.add_argument('-y', '--height', required=True, type=int, 
                        help=help('-y'))
    parser.add_argument('-a', '--algo', required=True, type=str,
                        help=help('-a'))
    parser.add_argument('-t', '--time-segmentation', required=False, 
                        default=False, type=int, help=help('-t'))
    parser.add_argument('-f', '--fps', required=False, 
                        default=None, type=float, help=help('-f'))

    # Read parameters
    args = parser.parse_args()

    # Modify parameters according to needs
    args.time_segmentation = bool(args.time_segmentation)
    
    return args


def validate_cmdline_params(args):
    """
    @brief Input directory must exist and output must not.
    """
    if not os.path.isfile(args.input) and not os.path.isdir(args.input):
        raise RuntimeError('[ERROR] Input file does not exist.')
    if os.path.isfile(args.output) or os.path.isdir(args.output):
        raise RuntimeError('[ERROR] Output file already exists.')
    return args


def process_video(input_path, output_path, args):
    # Create video summariser
    vidsum = videosum.VideoSummariser(args.algo, args.nframes, 
                                      args.width, args.height, 
                                      time_segmentation=args.time_segmentation,
                                      fps=args.fps)

    # Summarise video
    im = vidsum.summarise(input_path)

    # Save summary to the output folder
    cv2.imwrite(output_path, im, [int(cv2.IMWRITE_JPEG_QUALITY), 100])


def main():
    # Read command line parameters
    args = parse_cmdline_params()
    validate_cmdline_params(args)
    
    # Check whether the input is a file or a folder of videos
    if os.path.isfile(args.input):
        # Create video summariser
        vidsum = videosum.VideoSummariser(args.algo, args.nframes, 
                                          args.width, args.height, 
                                          time_segmentation=args.time_segmentation,
                                          fps=args.fps)
        # The input is a file
        tic = time.time()
        im = vidsum.summarise(args.input)
        toc = time.time()
        print("[INFO] Video summarised in {} seconds.".format(toc - tic))
        cv2.imwrite(args.output, im, [int(cv2.IMWRITE_JPEG_QUALITY), 100])
    else:
        # The input is a folder of videos
        input_dir = args.input
        output_dir = args.output
        
        # Create output folder
        os.mkdir(output_dir)

        # Gather the list of video filenames inside the folder
        videos = [x for x in os.listdir(input_dir) if x.endswith('.mp4')]
        
        # Build data input ready for batch processing
        data_inputs = []
        for v in videos:
            input_path = os.path.join(input_dir, v)
            output_path = os.path.join(output_dir, os.path.splitext(v)[0] + '.jpg')
            data_inputs.append((input_path, output_path, args))
        
        # Run batch processing
        processes = 2 * multiprocessing.cpu_count()
        pool = multiprocessing.Pool(processes=processes)
        pool.starmap(process_video, data_inputs)
        #for input_path, output_path, args in data_input:
        #    print("[INFO] Processing {} ...".format(input_path))
        #    process_video(input_path, output_path, args)


if __name__ == '__main__':
    main()
